@@ -45,6 +45,7 @@ typedef struct {
     void *hook_data;
     char* temp_dir;
     request_rec *r;
+    int nargs;
 } ApacheRequest;
 
 struct ApacheUpload {
@@ -97,6 +98,8 @@ char *ApacheRequest_params_as_string(ApacheRequest *req, const char *key);
 int ApacheRequest___parse(ApacheRequest *req);
 #define ApacheRequest_parse(req) \
     (req->status = req->parsed ? req->status : ApacheRequest___parse(req)) 
+table *ApacheRequest_query_params(ApacheRequest *req, ap_pool *p);
+table *ApacheRequest_post_params(ApacheRequest *req, ap_pool *p);
 
 FILE *ApacheRequest_tmpfile(ApacheRequest *req, ApacheUpload *upload);
 ApacheUpload *ApacheUpload_new(ApacheRequest *req);
@@ -116,6 +119,7 @@ ap_table_get(upload->info, key)
 ApacheUpload_info(upload, "Content-Type")
 
 #define ApacheRequest_set_post_max(req, max) (req->post_max = max)
+#define ApacheRequest_set_temp_dir(req, dir) (req->temp_dir = dir)
 
 char *ApacheUtil_expires(pool *p, char *time_str, int type);
 #define EXPIRES_HTTP   1
